import sys
import os
import readchar

def clear_last_lines(num_lines):
    for _ in range(num_lines):
        # Move cursor up one line
        sys.stdout.write("\033[F")
        # Clear the line
        sys.stdout.write("\033[K")

def clear_console():
    os.system("cls" if os.name == "nt" else "clear")

def draw_menu(options, selected_idx):
    #clear_console()
    for idx, option in enumerate(options):
        prefix = "> " if idx == selected_idx else "  "
        print(f"{prefix}{option}")

def selector_menu(options):
    selected_idx = 0
    while True:
        draw_menu(options, selected_idx)
        key = readchar.readkey()
        if key == readchar.key.UP and selected_idx > 0:
            selected_idx -= 1
            clear_last_lines(len(options))
        elif key == readchar.key.DOWN and selected_idx < len(options) - 1:
            selected_idx += 1
            clear_last_lines(len(options))
        elif key == readchar.key.ENTER:
            return selected_idx


import sys
import os
import readchar
import json
from http.server import SimpleHTTPRequestHandler, HTTPServer
import webbrowser
import importlib.resources as pkg_resources

from .constants import version
from .testing_agent import start_testing
import threading
import asyncio
import subprocess
import signal

def initialize():
    if os.path.isfile(".yumevalidator.json"):
        overwrite = input("File .yumevalidator.json already exists. Do you want to overwrite it? (y/n): ")
        if overwrite.lower() != 'y':
            return "Initialization cancelled"
    
    figma_token = input("Please enter your Figma token: ")
    figma_key = input("Please enter the relevant figma board key: ")
    openai_api_key = input("Please enter your OpenAI API key:")
    website_url = input("Please enter the website URL you want to validate: ")

    with open(".yumevalidator.json", "w") as f:
        json.dump({
            "figma_token": figma_token,
            "figma_key": figma_key,
            "openai_api_key": openai_api_key,
            "website_url": website_url
        }, f, indent=4)
    
    with open(".gitignore") as f:
        if ".yumevalidator.json" not in f.read():
            with open(".gitignore", "a") as f:
                f.write("\n.yumevalidator.json")
        
    return "initialzing project"

def host_report_data():
    class CORSRequestHandler(SimpleHTTPRequestHandler):
        def end_headers(self):
            self.send_header('Access-Control-Allow-Origin', '*')
            self.send_header('Access-Control-Allow-Methods', 'GET, POST, OPTIONS')
            self.send_header('Access-Control-Allow-Headers', 'Content-Type')
            super().end_headers()

    print("Current directory:", os.getcwd())
    host_current_directory = os.getcwd()
    os.chdir(host_current_directory + "/.yumevalidator")
    host_handler = CORSRequestHandler
    host_httpd = HTTPServer(('localhost', 9001), host_handler)
    print("Serving on http://localhost:9001 server")
    host_httpd.serve_forever()

def display_reporting_screen():
    with pkg_resources.path(__package__, 'public') as public_dir:
        print(public_dir)
        os.chdir(public_dir)
        handler = SimpleHTTPRequestHandler
        httpd = HTTPServer(('localhost', 9000), handler)
        print("Serving on http://localhost:9000")
        webbrowser.open("http://localhost:9000")
        httpd.serve_forever()

def main():
    args = sys.argv[1:]
    if not args:
        raise ValueError("No command line arguments provided.")
    
    if args == ["version"]:
        return "Version " + version

    if args == ["init"]:
        return initialize()
    
    if args == ["test"]:
        start_testing()
        report_process = subprocess.Popen(["yumevalidator", "report"])
        report_process.wait()
        return

    if args == ["report-server"]:
        return host_report_data()
    
    if args == ["report-client"]:
        return display_reporting_screen()

    if args == ["report"]:
        server_process = subprocess.Popen(["yumevalidator", "report-server"])
        client_process = subprocess.Popen(["yumevalidator", "report-client"])

        def cleanup():
            server_process.terminate()
            client_process.terminate()
            server_process.wait()
            client_process.wait()

        signal.signal(signal.SIGINT, cleanup)
        signal.signal(signal.SIGTERM, cleanup)
        signal.signal(signal.SIGHUP, cleanup)
        signal.signal(signal.SIGQUIT, cleanup)

        server_thread = threading.Thread(target=server_process.wait)
        client_thread = threading.Thread(target=client_process.wait)
        server_thread.start()
        client_thread.start()
        server_thread.join()
        client_thread.join()
        return
    
    if args == ["test_host"]:
        return host_report_data()
    
    print("Command line arguments:", args)

if __name__ == "__main__":
    main() 


# This function is responsible for handling the retrieval of figma-related data from the API
# Functionalities include:
# - Describe the Figma Webpage
# - Generate a list of all interactable elements
# - Obtain screenshot from Figma

import requests
import json
import asyncio
from io import BytesIO

from smolagents import OpenAIServerModel, tool
from openai import OpenAI
from PIL import Image

from .constants import *

import base64
import os

figma_contents = []
figma_pages = []
interactable_elements_body = []

figma_file_structure = []

def get_updated_figma_pages():
    return figma_pages

async def figma_get_all_pages():
    """download and update contents from figma into a variable to be processed"""
    with open(".yumevalidator.json") as f:
        global figma_contents
        global figma_pages
        global figma_file_structure
        
        config = json.load(f)
        figma_token = config["figma_token"]
        figma_key = config["figma_key"]
        response = requests.get(f"https://api.figma.com/v1/files/{figma_key}", headers={"X-Figma-Token": figma_token})
        data = response.json()
        document_children = data["document"]["children"][0]["children"]
        print(len(document_children))
        figma_contents = document_children
        figma_pages = [page["id"] for page in document_children]
        
        def build_file_structure(node, depth=0, max_depth=4):
            if depth > max_depth:
                return None
            structure = {
            "name": node["name"],
            "id": node["id"],
            "children": []
            }
            if "children" in node:
                for child in node["children"]:
                    child_structure = build_file_structure(child, depth + 1, max_depth)
                    if child_structure:
                        structure["children"].append(child_structure)
            return structure

        figma_file_structure = [build_file_structure(page) for page in document_children]

def get_figma_file_structure():
    return figma_file_structure

async def figma_print_target(target_index):
    """
    print out a specific page from figma and save it as a png file with the name {target_page_id}.png
    """
    with open(".yumevalidator.json") as f:
        config = json.load(f)
        figma_token = config["figma_token"]
        figma_key = config["figma_key"]
        response = requests.get(f"https://api.figma.com/v1/images/{figma_key}?ids={target_index}", headers={"X-Figma-Token": figma_token})
        image_url = response.json()["images"][target_index]
        image_response = requests.get(image_url)
        if not os.path.exists(".yumevalidator"):
            os.makedirs(".yumevalidator")
        with open(f".yumevalidator/{target_index}.png", "wb") as img_file:
            img_file.write(image_response.content)
            return f"{target_index}.png"

def figma_get_page_name(page_index):
    # get the name of the page from the figma contents
    target_page = [node for node in figma_contents if node["id"] == page_index][0]
    return target_page["name"]

def figma_get_all_interactable_elements_from_node(target_page_index):
    # target_page is a json file that contains information about a figma node
    # repeatable parse the target_page to find all interactable elements
    # interactable elements have non-empty "interactions" field
    interactable_elements = []

    def find_interactable_elements(node):
        global interactable_elements_body
        stack = [node]
        while stack:
            current_node = stack.pop()
            if "children" in current_node:
                stack.extend(current_node["children"])
            if "interactions" in current_node and current_node["interactions"]:
                interactable_elements.append(current_node["id"])
                interactable_elements_body.append(current_node)

    target_base_node = [node for node in figma_contents if node["id"] == target_page_index][0]
    find_interactable_elements(target_base_node)
    return interactable_elements

async def figma_describe_screen(screen_image):
    # describe what the screen is about to be fed into the next agent
    with open(".yumevalidator.json") as f:
        config = json.load(f)
        client = OpenAI(api_key=config["claude_api_key"])
        with open(f".yumevalidator/{screen_image}", "rb") as image_file:
            encoded_string = base64.b64encode(image_file.read()).decode("utf-8")
            img_url = f"data:image/png;base64,{encoded_string}"
            response = client.chat.completions.create(
                model=main_model,
                messages=[
                    {
                        "role": "user",
                        "content": [
                            {"type": "text", "text": "This is a screen of a website. Please describe what you see."},
                            {
                                "type": "image_url",
                                "image_url": {"url": f"{img_url}"},
                            },
                        ],
                    }
                ],
            )
            return response.choices[0].message.content

@tool
def figma_get_interaction_target(element_id: str) -> str:
    """
    for a specific element, return the destination interaction target
    Args:
        element_id: the id of the element in question
    Returns:
        the target id of the interaction to be used in get_image to know how it looks like
    """
    target_base_node = [node for node in interactable_elements_body if node["id"] == element_id][0]
    if "interactions" in target_base_node:
        return target_base_node["interactions"][0]["actions"][0]["destinationId"]
    pass

if __name__ == "__main__":
    pages = asyncio.run(figma_get_all_pages())
    #print(json.dumps(pages[0], indent=4))
    screen_file_name = asyncio.run(figma_print_target(figma_pages[0]))
    figma_get_all_interactable_elements_from_node(figma_pages[0])
    #print(json.dump(interactable_elements_body, indent=4)
    print(screen_file_name)
    print(asyncio.run(figma_describe_screen(
        screen_file_name
    )))
    print(asyncio.run(figma_get_interaction_target("1:6")))
    pass
#print(asyncio.run(figma_get_all_interactable_elements_from_node(pages[0])))
#figma_print_image(pages[0]["id"])


        



from io import BytesIO
from time import sleep

import helium
import json
import asyncio

from dotenv import load_dotenv
from PIL import Image
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

from smolagents import CodeAgent, tool
from smolagents.agents import ActionStep
from smolagents import CodeAgent, OpenAIServerModel

from .constants import main_model, helium_instructions
from .figma_functions import *
from .difference_image_generator import diff_img

chrome_options = webdriver.ChromeOptions()
chrome_options.add_argument("--force-device-scale-factor=1")
chrome_options.add_argument("--window-size=1000,1350")
chrome_options.add_argument("--disable-pdf-viewer")
chrome_options.add_argument("--window-position=0,0")

global running_visual_testing_agent
running_visual_testing_agent: CodeAgent = None

global buffered_figma_image
buffered_figma_image = None

global screenshot_image
screenshot_image = None

global visual_testing_summary
visual_testing_summary = []

global current_working_list
current_working_list = []

global current_wrong_list
current_wrong_list = []

global current_recommendations_list
current_recommendation_list = []

global current_page_name
current_page_name = ""

global current_page_id
current_page_id = ""

def start_browser():
    global driver
    driver = helium.start_chrome(headless=False, options=chrome_options)

@tool
def save_screenshot(page_id: str) -> bool:
    """
    description: take a screenshot of the current page and save it to the memory with figma design as the context for comparison
    Args:
        page_id: the id of the page in question
    returns:
        true: screenshot has been saved successfully
        false: screenshot saving failed
    """
    sleep(1.0)  # Let JavaScript animations happen before taking the screenshot
    driver = helium.get_driver()
    if driver is not None:

        global screenshot_image
        global buffered_figma_image

        screenshot_image = (Image.open(BytesIO(driver.get_screenshot_as_png())))
        buffered_figma_image = Image.open(f".yumevalidator/{page_id}.png")

        if not os.path.exists(".yumevalidator"):
            os.makedirs(".yumevalidator")

        screenshot_image.save(f".yumevalidator/screenshot_{page_id}.png")


        return True
    return False

def add_image_to_observation(memory_step: ActionStep, agent: CodeAgent) -> None:
    
    global screenshot_image
    global buffered_figma_image
    global current_page_id

    if screenshot_image is not None and buffered_figma_image is not None:
        #diff_img(current_page_id)
        #memory_step.observations_images = [Image.open(f".yumevalidator/diff_{current_page_id}.png").copy()]
        memory_step.observations_images = [screenshot_image.copy()]
        memory_step.observations_images.append(buffered_figma_image.copy())
        screenshot_image.close()
        buffered_figma_image.close()
        screenshot_image = None
        buffered_figma_image = None

def figma_get_image(element_id: str) -> bool:
    """
     obtain how a particular element look like on the figma design whether if it is a specific element or an entire page
     It will be passed into memory during the next step

     Args:
        element_id: the id of the element in question
    Returns:
        true: image has been retrieved successfully
        false: image retrieval failed
     """
    
    with open(f".yumevalidator/{element_id}.png", "rb") as img_file:
        global buffered_figma_image
        buffered_figma_image = Image.open(BytesIO(img_file.read())).copy()
        print(buffered_figma_image.size)
        return True

@tool
def end_visual_testing(page_title: str, page_id: str) -> str:
    """
    End the testing of the current page
    Args:
        page_title: the human-readable name of the page
        page_id: the id of the page
    Returns:
        status
    """
    global visual_testing_summary
    visual_testing_summary.append(
        {
            "page_title": page_title,
            "page_id": page_id,
            "working": current_working_list,
            "wrong": current_wrong_list,
            "recommendations": current_recommendation_list,
        }
    )
    return "Testing of the current page can be ended successfully"

@tool
def add_to_working_list(title: str, description: str) -> str:
    """
    Record which part of the visual testing is working
    Args:
        title: the title of the working part
        description: the description of the working part
    """
    global current_working_list
    current_working_list.append({
            "title": title,
            "description": description
        })
    return "Working part has been added successfully"

@tool
def add_to_wrong_list(title: str, description: str, target_element_id: str) -> str:
    """
    Record which part of the visual testing is wrong
    Args:
        title: the title of the wrong part
        description: the description of the wrong part
        target_element_id: the id of the element that is wrong (it is the non-human readable token)
    """
    global current_wrong_list
    current_wrong_list.append({
            "title": title,
            "description": description,
            "target_element": target_element_id
        })
    return "Wrong part has been added successfully"

@tool
def add_to_recommendation_list(title: str, description: str) -> str:
    """
    Record the suggestion you would give to fix the issues faced
    Args:
        title: the title of the recommendation
        description: the description of the recommendation
    """
    global current_recommendation_list
    current_recommendation_list.append(
        {
            "title": title,
            "description": description
        })
    return "Recommendation has been added successfully"

def start_visual_testing_agent():
    
    global visual_testing_summary
    global current_working_list
    global current_wrong_list
    global current_recommendation_list
    global current_page_name
    global current_page_id

    visual_testing_summary = []
    current_working_list = []
    current_wrong_list = []
    current_recommendation_list = []
    current_page_name = ""
    current_page_id = ""

    start_browser()
    asyncio.run(figma_get_all_pages())
    with open(".yumevalidator.json") as f:
        global visual_test_untested_user_interface
        visual_test_untested_user_interface = get_updated_figma_pages()
        config = json.load(f)
        model = OpenAIServerModel(main_model, "https://api.openai.com/v1", api_key=config["openai_api_key"])

        for page in visual_test_untested_user_interface:
            current_working_list = []
            current_wrong_list = []
            current_recommendation_list = []

            current_page_id = page
            current_page_name = figma_get_page_name(page)
            asyncio.run(figma_print_target(page))
            interactable_nodes = figma_get_all_interactable_elements_from_node(page)
            #page_description = asyncio.run(figma_describe_screen(f"{page}.png"))
            page_description = ""
            for interactable_node in interactable_nodes:
                asyncio.run(figma_print_target(interactable_node))
            
            # Create the agent
            global running_visual_testing_agent
            running_visual_testing_agent = CodeAgent(
                tools=[save_screenshot, end_visual_testing, add_to_recommendation_list, add_to_wrong_list, add_to_working_list],
                model=model,
                additional_authorized_imports=["helium"],
                step_callbacks=[add_image_to_observation],
                max_steps=8,
                verbosity_level=2,
            )

            # Import helium for the agent
            running_visual_testing_agent.python_executor("from helium import *", running_visual_testing_agent.state)
            #running_visual_testing_agent.python_executor("def print(*args, **kwargs): pass", running_visual_testing_agent.state)
            
            execution_request =f"""
            Your objective is to test the visual fidelity of the react project by looking at the Figma Design Image, and the actual website. 3 or more Context Images will be given, one will the figma design file, one will be the implementation rendered webpage and one will be the fidelity difference image

            The title of the page that needs to be tested at the moment is
            {figma_get_page_name(page)}

            The id of the current page is:
            {page}

            The base website url is:
            {
                config["website_url"]
            }
            Here are the available paths of this website, you can visit them by {config["website_url"]}/[path]'
            ```
            {
                os.listdir("src/app")
            }
            ````<end_code>

            Take these steps step by step to test the interaction functionality, and mention which step you are taking. ONLY test the ones that have been mentioned, just because an element is visible, it is not meant to be tested, because it is still under production. Strictly, do not do multiple steps at one time, because it will crash the browser.
            1. Visit the Website
            2. The path of current website is {driver.current_url}, if it is not the same as {figma_get_page_name(page)}, then switch to the correct page. Note that, the image of the page attached does not represent the current page. Homepage typically lives on "/home" or "/index" or "/"
            3. Take a screenshot of the website
            4. Look at the attached figma design, and website screenshot.
            5. Describe what is right, and record it with add_to_working_list tool (there can be muliples)
            6. Describe what is wrong, and record it with add_to_wrong_list tool (there can be multiples)
            7. Describe what can be improved, and record it with add_to_recommendation_list tool (there can be multiples)
            5. End the test

            If the website is entirely white, it is most likely that you forgot to visit the website after opening the browser. Do not do multiple actions in one settings, because the browser might not load as fast as you think.

            Sometimes, the starting webpage might be different from the actual page you should be trying to test. Switch to the relevant website by using the following code:
            Code:
            ```py
            go_to("{config["website_url"]}/" + [[current target potential page]]")
            ```<end_code>

            This is how the figma file structure looks like. If needed find the relevant elements in the figma file and get their id for whatever use cases
            \"\"\"
            {get_figma_file_structure()}
            \"\"\"

            You can use the tools to determine how the specified element looks like on the page and what is the result of interacting with that particular element.
            
            Even if I say there is "Error in Code Parsing", do not worry about it, because it is just a warning, and it will not affect the testing process.
            
            """

            #execution_request = """
            #Please navigate to https://en.wikipedia.org/wiki/Chicago and give me a sentence containing the word "1992" that mentions a construction accident.
            #"""
            try:
                running_task = running_visual_testing_agent.run(execution_request + helium_instructions)
            except TypeError:
                print("completed")
        
        if not os.path.exists(".yumevalidator"):
            os.makedirs(".yumevalidator")
        with open(".yumevalidator/visual_testing_results.json", "w") as outfile:
            json.dump(visual_testing_summary, outfile, indent=4)


from .visual_testing_agent import start_visual_testing_agent
from .functional_testing_agent import start_functional_testing_agent

def start_testing():
    start_visual_testing_agent()
    start_functional_testing_agent()




# Credits to James

import os
import json
import base64
import subprocess

global page_id
page_id = None

def save_base64_image(base64_string, filename):
    """Decodes a base64 image string and saves it as a file."""
    try:
        header, encoded = base64_string.split(",", 1)  # Split header and encoded data
        img_data = base64.b64decode(encoded)  # Decode base64
        with open(filename, "wb") as img_file:
            img_file.write(img_data)
        print(f"Saved: {filename}")
    except Exception as e:
        print(f"Error saving {filename}: {e}")

def process_json(json_path):
    """Reads JSON file, finds base64 images, and saves them with a sequential filename."""
    with open(json_path, "r", encoding="utf-8") as file:
        data = json.load(file)

    img_counter = 0
    img_dir = ".yumevalidator/current_fi/extracted_images"
    os.makedirs(img_dir, exist_ok=True)  # Create directory for images

    def extract_and_save(obj, key, path="root"):
        nonlocal img_counter
        if isinstance(obj, dict):
            for k, v in obj.items():
                current_path = f"{path}.{k}"
                # Check if the value is a base64-encoded PNG image
                if k == key and isinstance(v, str) and v.startswith("data:image/png;base64,"):
                    filename = os.path.join(img_dir, f"image_{img_counter}.png")
                    save_base64_image(v, filename)
                    print(f"Found image at: {current_path}, saved as: {filename}")
                    img_counter += 1
                else:
                    extract_and_save(v, key, current_path)
        elif isinstance(obj, list):
            for i, item in enumerate(obj):
                extract_and_save(item, key, f"{path}[{i}]")

    # Look for the base64 image data in fields named "url"
    extract_and_save(data, "url")
    print(f"Total images saved: {img_counter}")
    # Move the images to another location
    new_img_dir = f".yumevalidator/diff_{page_id}.png"
    os.makedirs(new_img_dir, exist_ok=True)
    for img_file in os.listdir(img_dir):
        if img_file == "image_2.png":
            src_path = os.path.join(img_dir, img_file)
            dst_path = os.path.join(new_img_dir, img_file)
            os.rename(src_path, dst_path)
            break
    print(f"Images moved to: {new_img_dir}")
    

def diff_img(new_page_id):
    """
    Compare the image of the page with the image of the page in the previous commit
    """
    global page_id

    install_scoopy = subprocess.Popen(["npm", "install", "-g", "@animaapp/scooby-cli"])
    install_scoopy.wait()

    fidelity_generator = subprocess.Popen(["scooby", "fidelity", "--name", "\".yumevalidator/current_fi\"", "--expected", f".yumevalidator/expected", "--actual", f".yumevalidator/actual", "--file-type=png"])
    fidelity_generator.wait()
    subprocess.run(["unzip", ".yumevalidator/current_fi.zip", "-d", ".yumevalidator/current_fi"])
    process_json(".yumevalidator/current_fi/report.json")
    page_id = new_page_id

    pass

def main():
    diff_img("1:247")


version="1.0"

main_model="gpt-4o-mini"
instruct_openai_model="o1-mini"

helium_instructions = """
    You can use helium to access websites. Don't bother about the helium driver, it's already managed.
    We've already ran "from helium import *"
    Then you can go to pages!
    Code:
    ```py
    go_to('github.com/trending')
    ```<end_code>

    You can directly click clickable elements by inputting the text that appears on them.
    Code:
    ```py
    click("Top products")
    ```<end_code>

    If it's a link:
    Code:
    ```py
    click(Link("Top products"))
    ```<end_code>

    If you try to interact with an element and it's not found, you'll get a LookupError.
    In general stop your action after each button click to see what happens on your screenshot.
    Never try to login in a page.

    To scroll up or down, use scroll_down or scroll_up with as an argument the number of pixels to scroll from.
    Code:
    ```py
    scroll_down(num_pixels=1200) # This will scroll one viewport down
    ```<end_code>

    When you have pop-ups with a cross icon to close, don't try to click the close icon by finding its element or targeting an 'X' element (this most often fails).
    Just use your built-in tool `close_popups` to close them:
    Code:
    ```py
    close_popups()
    ```<end_code>

    You can use .exists() to check for the existence of an element. For example:
    Code:
    ```py
    if Text('Accept cookies?').exists():
        click('I accept')
    ```<end_code>
    """


import os
from http.server import SimpleHTTPRequestHandler, HTTPServer

class CORSRequestHandler(SimpleHTTPRequestHandler):
    def end_headers(self):
        self.send_header('Access-Control-Allow-Origin', '*')
        self.send_header('Access-Control-Allow-Methods', 'GET, POST, OPTIONS')
        self.send_header('Access-Control-Allow-Headers', 'Content-Type')
        super().end_headers()

os.chdir("sample_json")
handler = CORSRequestHandler
httpd = HTTPServer(('localhost', 9001), handler)
httpd.serve_forever()


from setuptools import setup, find_packages

setup(
    name="yumevalidator",  # Change to your package name
    version="0.2.0",
    packages=find_packages(),
    include_package_data=True,
    package_data={
        '': ['public/*'],
    },
    install_requires=[
        'readchar',
        'smolagents',
        'selenium',
        'helium',
        'pillow',
        'openai',
        'cors'
    ],  # Add dependencies if needed
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "yumevalidator=src.main:main", # Command name -> Function to run
        ],
    },
    author="Victor Mak",
    author_email="contact@htetaung.com",
    description="A GenAI-based UI Validation tool for Deriv AI Hackathon in 2025 Feburary",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/yumevalidator/cli",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


from io import BytesIO
from time import sleep

import helium
import json
import asyncio

from dotenv import load_dotenv
from PIL import Image
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

from smolagents import CodeAgent, tool
from smolagents.agents import ActionStep
from smolagents import CodeAgent, OpenAIServerModel

from .constants import main_model, helium_instructions
from .figma_functions import *
# Configure Chrome options
chrome_options = webdriver.ChromeOptions()
chrome_options.add_argument("--force-device-scale-factor=1")
chrome_options.add_argument("--window-size=1000,1350")
chrome_options.add_argument("--disable-pdf-viewer")
chrome_options.add_argument("--window-position=0,0")

global driver
global buffered_figma_image
global functional_testing_summary

buffered_figma_image = None
functional_testing_summary = []

def start_browser():
    global driver
    driver = helium.start_chrome(headless=False, options=chrome_options)

functional_test_untested_user_interface = []

# Set up screenshot callback
def save_screenshot(memory_step: ActionStep, agent: CodeAgent) -> None:
    sleep(1.0)  # Let JavaScript animations happen before taking the screenshot
    driver = helium.get_driver()
    current_step = memory_step.step_number
    if driver is not None:
        for previous_memory_step in agent.memory.steps:  # Remove previous screenshots for lean processing
            if isinstance(previous_memory_step, ActionStep) and previous_memory_step.step_number <= current_step - 2:
                previous_memory_step.observations_images = None
        png_bytes = driver.get_screenshot_as_png()

        image = Image.open(BytesIO(png_bytes))
        print(f"Captured a browser screenshot: {image.size} pixels")
        memory_step.observations_images = [image.copy()]  # Create a copy to ensure it persists

    # Update observations with current URL
    url_info = f"Current url: {driver.current_url}"
    memory_step.observations = (
        "current_website_path=" + url_info
    )

def add_buffered_observations(memory_step: ActionStep, agent: CodeAgent) -> None:
    global buffered_figma_image
    if buffered_figma_image is not None:
        print("Adding buffered Figma image to observations")
        memory_step.observations_images.append(buffered_figma_image.copy())
        buffered_figma_image.close()
        buffered_figma_image = None

@tool
def figma_get_image(element_id: str) -> bool:
    """
     obtain how a particular element look like on the figma design whether if it is a specific element or an entire page
     It will be passed into memory during the next step

     Args:
        element_id: the id of the element in question
    Returns:
        true: image has been retrieved successfully
        false: image retrieval failed
     """
    
    with open(f".yumevalidator/{element_id}.png", "rb") as img_file:
        global buffered_figma_image
        buffered_figma_image = Image.open(BytesIO(img_file.read())).copy()
        print(buffered_figma_image.size)
        return True

@tool
def search_item_ctrl_f(text: str, nth_result: int = 1) -> str:
    """
    Searches for text on the current page via Ctrl + F and jumps to the nth occurrence.
    Args:
        text: The text to search for
        nth_result: Which occurrence to jump to (default: 1)
    """
    elements = driver.find_elements(By.XPATH, f"//*[contains(text(), '{text}')]")
    if nth_result > len(elements):
        raise Exception(f"Match n°{nth_result} not found (only {len(elements)} matches found)")
    result = f"Found {len(elements)} matches for '{text}'."
    elem = elements[nth_result - 1]
    driver.execute_script("arguments[0].scrollIntoView(true);", elem)
    result += f"Focused on element {nth_result} of {len(elements)}"
    return result

@tool
def go_back() -> None:
    """Goes back to previous page."""
    driver.back()

@tool
def close_popups() -> str:
    """
    Closes any visible modal or pop-up on the page. Use this to dismiss pop-up windows!
    This does not work on cookie consent banners.
    """
    webdriver.ActionChains(driver).send_keys(Keys.ESCAPE).perform()


@tool
def end_testing_page(page_title: str, is_there_error: bool, testing_description: str) -> str:
    """
    End the testing of the current page
    Args:
        page_title: human-readable name of the page that has beent tested",
        is_there_error: whether if there is an error during the testing
        testing_description: the description of the testing, what has been tested in a organized list, with errors listed out properly
    Returns:
        status
    """
    global functional_testing_summary
    functional_testing_summary.append({
        "page_title": page_title,
        "succeeded": not is_there_error, 
        "description": testing_description
        })
    return "Testing of the current page can be ended successfully"


def start_functional_testing_agent():
    
    global functional_testing_summary
    functional_testing_summary = []
    start_browser()
    asyncio.run(figma_get_all_pages())
    with open(".yumevalidator.json") as f:
        global functional_test_untested_user_interface
        functional_test_untested_user_interface = get_updated_figma_pages()
        config = json.load(f)
        model = OpenAIServerModel(main_model, "https://api.openai.com/v1", api_key=config["openai_api_key"])

        for page in functional_test_untested_user_interface:
            asyncio.run(figma_print_target(page))
            interactable_nodes = figma_get_all_interactable_elements_from_node(page)
            #page_description = asyncio.run(figma_describe_screen(f"{page}.png"))
            page_description = ""
            for interactable_node in interactable_nodes:
                asyncio.run(figma_print_target(interactable_node))
            
            # Create the agent
            agent = CodeAgent(
                tools=[close_popups, search_item_ctrl_f, figma_get_image, figma_get_interaction_target, end_testing_page],
                model=model,
                additional_authorized_imports=["helium"],
                step_callbacks=[save_screenshot, add_buffered_observations],
                max_steps=20,
                verbosity_level=2,
            )
            # Import helium for the agent
            agent.python_executor("from helium import *", agent.state)
            
            execution_request =f"""
            Your objective is to test the functionality of the react project by looking at the Figma Design Image, and interacting with the specified elements on the page.
            It might lead you to a different webpage, but you should always try to go back to the original webpage to finish testing the elements on that page.

            The title of the page that needs to be tested at the moment is
            {figma_get_page_name(page)}


            The base website url is:
            {
                config["website_url"]
            }

            Here are the available paths of this website, you can visit them by {config["website_url"]}/[path]'
            ```
            {
                os.listdir("src/app")
            }
            ````<end_code>

            Take these steps step by step to test the interaction functionality. ONLY test the ones that have been mentioned, just because an element is visible, it is not meant to be tested, because it is still under production. Strictly, do not do multiple steps at one time, because it will crash the browser.
            1. Visit the Website
            2. The path of current website is {driver.current_url}, if it is not the same as {figma_get_page_name(page)}, then switch to the correct page. Note that, the image of the page attached does not represent the current page.
            3. Look into the interactable_nodes, one after one, do not look into random ones, because they have not been implemented
            4. Use the tool to retrieve the image of the element that is currently being processed
            4. Look at where they lead to
            5. Interact with the element
            6. Check whether if they go to the correct page as intended
            7. Go back to original webpage
            8. After every elements has been checked, end the test for this screen

            If the website is entirely white, it is most likely that you forgot to visit the website after opening the browser. Do not do multiple actions in one settings, because the browser might not load as fast as you think.

            Sometimes, the starting webpage might be different from the actual page you should be trying to test. Switch to the relevant website by using the following code:
            Code:
            ```py
            go_to("{config["website_url"]}/" + [[current target potential page]]")
            ````<end_code>

            You can use the tools to determine how the specified element looks like on the page and what is the result of interacting with that particular element.

            Here is the array of interactable elements on the page in a python array
            Code:
            ```py
            {interactable_nodes}
            ````<end_code>
            
            Here is the description of the current page:
            '
            {page_description}
            '


            #execution_request = """
            #Please navigate to https://en.wikipedia.org/wiki/Chicago and give me a sentence containing the word "1992" that mentions a construction accident.
            #"""
            try:
                agent.run(execution_request + helium_instructions)
            except TypeError:
                print("completed")
    if not os.path.exists(".yumevalidator"):
        os.makedirs(".yumevalidator")
    with open(".yumevalidator/visual_testing_results.json", "w") as outfile:
        json.dump(functional_testing_summary, outfile, indent=4)

if __name__ == "__main__":        
    start_functional_testing_agent()